"""Defines the LMFit tab and the LMFitResult tab.

This file defines two similar types of tabs: the LMFit and the
LMFitResult tab as well as a common base class LMFitBaseTab.
"""
# Python Imports
import logging

# Third Party Imports
import numpy as np

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import QDir, pyqtSignal, Qt
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout

# Own Imports
from kmap import __directory__
from kmap.library.axis import Axis
from kmap.library.qwidgetsub import Tab
from kmap.model.lmfit_model import LMFitModel
from kmap.controller.dataslider import DataSlider
from kmap.controller.colormap import Colormap
from kmap.controller.crosshairannulus import CrosshairAnnulus
from kmap.controller.interpolation import LMFitInterpolation
from kmap.controller.lmfittree import LMFitTree, LMFitResultTree
from kmap.controller.lmfitresult import LMFitResult
from kmap.controller.lmfitoptions import LMFitOptions
from kmap.controller.lmfitorbitaloptions import LMFitOrbitalOptions
from kmap.config.config import config

# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/lmfittab.ui')
LMFitTab_UI, _ = uic.loadUiType(UI_file)


class LMFitBaseTab(Tab):

    def refresh_sliced_plot(self):

        index = self.slider.get_index()
        kmap = self.model.get_sliced_kmap(index)

        self.sliced_plot.plot(kmap)

    def refresh_selected_plot(self, param=None):

        ID = self.tree.get_selected_orbital_ID()

        if ID == -1:
            self.selected_plot.plot(None)

        else:
            kmap = self.model.get_orbital_kmap(ID, param)
            self.selected_plot.plot(kmap)

    def refresh_sum_plot(self, param=None):

        kmap = self.model.get_weighted_sum_kmap(param)
        self.sum_plot.plot(kmap)

        return kmap

    def refresh_residual_plot(self, param=None, weight_sum_data=None):

        index = self.slider.get_index()
        residual = self.model.get_residual(index, param, weight_sum_data)

        level = np.nanmax(np.absolute(residual.data))

        if config.get_key('pyqtgraph', 'keep_max_level_residual') == 'True':
            old_level = np.nanmax(np.absolute(self.residual_plot.get_levels()))

            if old_level != 1 and old_level > level:
                level = old_level

        self.residual_plot.plot(residual)

        self.residual_plot.set_levels([-level, level])

        self.update_chi2_label(weight_sum_data)

    def update_chi2_label(self, weight_sum_data=None):

        slice_index = self.slider.get_index()
        reduced_chi2 = self.model.get_reduced_chi2(
            slice_index, weight_sum_data)
        self.residual_label.setText('Residual (red. Chi^2: %.3E)'
                                    % reduced_chi2)

    def transpose(self, constant_axis):

        self.model.transpose(constant_axis)
        self.refresh_sliced_plot()
        self.refresh_residual_plot()

    def closeEvent(self, event):

        del self.model

        Tab.closeEvent(self, event)

    def _refresh_orbital_plots(self):

        self.refresh_selected_plot()
        kmap = self.refresh_sum_plot()
        self.refresh_residual_plot(weight_sum_data=kmap)

    def _setup(self):

        self.slider = DataSlider(self.model.sliced_data)
        self.crosshair = CrosshairAnnulus(self.residual_plot)
        self.colormap = Colormap(
            [self.sliced_plot, self.selected_plot, self.sum_plot])
        residual_colormap = Colormap(self.residual_plot)

        residual_colormap.set_colormap('blueAndRed')

    def _connect(self):

        self.crosshair.crosshair_changed.connect(self.crosshair.update_label)

        self.slider.slice_changed.connect(self.change_slice)
        self.slider.axis_changed.connect(self.change_slice)
        self.slider.tranpose_triggered.connect(self.transpose)

        self.tree.item_selected.connect(self.refresh_selected_plot)


class LMFitTab(LMFitBaseTab, LMFitTab_UI):

    fit_finished = pyqtSignal(list, list, dict)

    def __init__(self, sliced_data, orbitals):

        self.model = LMFitModel(sliced_data, orbitals)

        # Setup GUI
        super(LMFitTab, self).__init__()
        self.setupUi(self)

        self._setup()
        self._connect()

        self.refresh_sliced_plot()
        self.refresh_selected_plot()
        kmap = self.refresh_sum_plot()
        self.refresh_residual_plot(weight_sum_data=kmap)

    def get_title(self):

        return 'LM-Fit'

    def trigger_fit(self):

        try:
            results = self.model.fit()

        except ValueError as e:
            logging.getLogger('kmap').warning(str(e))
            self.lmfit_options.update_fit_button()
            
            return

        settings = self.model.get_settings()
        data = [self.model.sliced_data, self.model.orbitals]

        self.fit_finished.emit(data, results, settings)

    def change_slice(self):

        axis_index = self.slider.get_axis()
        slice_policy = self.lmfit_options.get_slice_policy()
        combined = True if slice_policy == 'all combined' else False
        slice_indices = (self.slider.get_index()
                         if slice_policy == 'only one' else 'all')

        self.model.set_slices(
            slice_indices, axis_index=axis_index, combined=combined)

        self.refresh_sliced_plot()
        self.refresh_residual_plot()

    def change_axis(self):

        axis = self.interpolation.get_axis()
        self.model.set_axis(axis)

        self.refresh_sliced_plot()
        self.refresh_selected_plot()
        kmap = self.refresh_sum_plot()
        self.refresh_residual_plot(weight_sum_data=kmap)

    def _change_slice_policy(self, slice_policy):

        axis = self.slider.get_axis()

        if slice_policy == 'all':
            self.model.set_slices('all', axis_index=axis, combined=False)

        elif slice_policy == 'only one':
            index = self.slider.get_index()
            self.model.set_slices([index], axis_index=axis, combined=False)

        else:
            self.model.set_slices('all', axis_index=axis, combined=True)

    def _change_region(self, *args):

        self.model.set_region(*args)
        self.refresh_residual_plot()

    def _change_background(self, *args):

        new_variables = self.model.set_background_equation(*args)
        for variable in new_variables:
            self.tree.add_equation_parameter(variable)

        self.refresh_sum_plot()
        self.refresh_residual_plot()

    def _setup(self):

        LMFitBaseTab._setup(self)

        self.orbital_options = LMFitOrbitalOptions()
        self.tree = LMFitTree(self.model.orbitals, self.model.parameters)
        self.interpolation = LMFitInterpolation()
        self.lmfit_options = LMFitOptions(self)

        self.change_axis()
        self.model.set_crosshair(self.crosshair.model)
        self._change_background(self.lmfit_options.get_background())

        layout = QVBoxLayout()
        layout.setContentsMargins(3, 3, 3, 3)
        layout.setSpacing(3)
        self.scroll_area.widget().setLayout(layout)
        layout.insertWidget(0, self.slider)
        layout.insertWidget(1, self.orbital_options)
        layout.insertWidget(2, self.interpolation)
        layout.insertWidget(3, self.lmfit_options)
        layout.insertWidget(4, self.colormap)
        layout.insertWidget(5, self.crosshair)

        self.layout.insertWidget(1, self.tree)

    def _connect(self):

        LMFitBaseTab._connect(self)

        self.interpolation.interpolation_changed.connect(self.change_axis)

        self.tree.value_changed.connect(self._refresh_orbital_plots)
        self.tree.vary_changed.connect(self.update_chi2_label)
        self.lmfit_options.background_changed.connect(self._change_background)
        self.lmfit_options.fit_triggered.connect(self.trigger_fit)
        self.lmfit_options.method_changed.connect(self.model.set_fit_method)
        self.lmfit_options.slice_policy_changed.connect(
            self._change_slice_policy)

        self.orbital_options.symmetrization_changed.connect(
            self.model.set_symmetrization)
        self.orbital_options.symmetrization_changed.connect(
            self._refresh_orbital_plots)
        self.orbital_options.polarization_changed.connect(
            self.model.set_polarization)
        self.orbital_options.polarization_changed.connect(
            self._refresh_orbital_plots)


class LMFitResultTab(LMFitBaseTab, LMFitTab_UI):

    open_plot_tab = pyqtSignal(list, list, Axis)

    def __init__(self, data, results, settings):

        self.results = results
        self.current_result = self.results[0]

        self.model = LMFitModel(*data)
        self.model.set_settings(settings)

        # Setup GUI
        super(LMFitResultTab, self).__init__()
        self.setupUi(self)

        self._setup()
        self._connect()

        self.refresh_sliced_plot()
        self.refresh_selected_plot()
        kmap = self.refresh_sum_plot()
        self.refresh_residual_plot(weight_sum_data=kmap)

    def get_title(self):

        return 'Results'

    def change_slice(self):

        index = self.slider.get_index()

        self.current_result = self.results[0] if len(
            self.results) == 1 else self.results[index]

        self.update_tree()
        self.result.result = self.current_result[1]
        self.refresh_sliced_plot()
        self._refresh_orbital_plots()

    def refresh_selected_plot(self):

        params = self.current_result[1].params
        super().refresh_selected_plot(params)

    def refresh_sum_plot(self):

        params = self.current_result[1].params
        return super().refresh_sum_plot(params)

    def refresh_residual_plot(self, weight_sum_data=None):

        params = self.current_result[1].params
        super().refresh_residual_plot(params, weight_sum_data)

    def update_tree(self):

        params = self.current_result[1].params
        self.tree.update_result(params)

    def print_result(self):

        report = self.result.get_fit_report()

        print(report)

    def print_covariance_matrix(self):

        cov_matrix = self.result.get_covariance_matrix()

        print(cov_matrix)

    def plot(self):

        results = [result[1] for result in self.results]
        orbitals = self.model.orbitals
        axis = self.model.sliced_data.axes[self.model.slice_policy[0]]
        self.open_plot_tab.emit(results, orbitals, axis)

    def _setup(self):

        LMFitBaseTab._setup(self)

        self.result = LMFitResult(self.current_result[1], self.model)
        self.tree = LMFitResultTree(
            self.model.orbitals, self.current_result[1].params,
            self.model.background_equation[1])
        self.crosshair._set_model(self.model.crosshair)

        layout = QVBoxLayout()
        self.scroll_area.widget().setLayout(layout)
        layout.insertWidget(0, self.slider)
        layout.insertWidget(1, self.result)
        layout.insertWidget(3, self.colormap)
        layout.insertWidget(4, self.crosshair)

        self.layout.insertWidget(1, self.tree)

    def _connect(self):

        self.result.print_triggered.connect(self.print_result)
        self.result.cov_matrix_requested.connect(self.print_covariance_matrix)
        self.result.plot_requested.connect(self.plot)

        LMFitBaseTab._connect(self)



# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtWidgets import QMainWindow
from PyQt5.QtCore import Qt, pyqtSignal, QDir

# Own Imports
from kmap import __directory__

# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/renametabwindow.ui')
RenameTabWindow_UI, _ = uic.loadUiType(UI_file)


class RenameTabWindow(QMainWindow, RenameTabWindow_UI):

    title_chosen = pyqtSignal(str)

    def __init__(self, *args, **kwargs):

        self.new_title = None

        # Setup GUI
        super(RenameTabWindow, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._setup()
        self._connect()

        self.show()

    def name_chosen(self):

        self.new_title = self.rename_tab_edit.text()

        self.close()

    def closeEvent(self, event):

        self.title_chosen.emit(self.new_title)
        self.deleteLater()
        event.accept()

    def _setup(self):

        self.setWindowFlags(Qt.CustomizeWindowHint | Qt.WindowTitleHint)

    def _connect(self):

        self.rename_tab_edit.returnPressed.connect(self.name_chosen)



# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import pyqtSignal, QDir

# Own Imports
from kmap import __directory__

# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/realplotoptions.ui')
RealPlotOptions_UI, _ = uic.loadUiType(UI_file)


class RealPlotOptions(QWidget, RealPlotOptions_UI):

    set_camera = pyqtSignal(int, int, int)
    show_grid_changed = pyqtSignal(int)
    show_mesh_changed = pyqtSignal(int)
    show_bonds_changed = pyqtSignal(int)
    show_photon_changed = pyqtSignal(int)
    show_hemisphere_changed = pyqtSignal(int)
    iso_val_changed = pyqtSignal()

    def __init__(self, plot_item):

        # Setup GUI
        super(RealPlotOptions, self).__init__()
        self.setupUi(self)
        self._connect()

    def reset_camera(self):

        distance = 75
        elevation = 90
        azimuth = -90
        self.set_camera.emit(distance, elevation, azimuth)

    def is_show_grid(self):

        return self.show_grid_checkbox.isChecked()

    def is_show_bonds(self):

        return self.show_bond_checkbox.isChecked()

    def is_show_hemisphere(self):

        return self.show_hemisphere_checkbox.isChecked()

    def is_show_photon(self):

        return self.show_photon_checkbox.isChecked()

    def is_show_mesh(self):

        return self.show_isosurface_checkbox.isChecked()

    def get_iso_val(self):

        return self.iso_spinbox.value()

    def _change_bonds_show(self, state):

        self.show_bonds_changed.emit(state)

    def _change_photon_show(self, state):

        self.show_photon_changed.emit(state)

    def _change_grid_show(self, state):

        self.show_grid_changed.emit(state)

    def _change_mesh_show(self, state):

        self.show_mesh_changed.emit(state)

    def _change_hemisphere_show(self, state):

        self.show_hemisphere_changed.emit(state)

    def _change_iso_val(self):

        self.iso_val_changed.emit()

    def _connect(self):

        self.reset_camera_button.clicked.connect(self.reset_camera)
        self.show_bond_checkbox.stateChanged.connect(self._change_bonds_show)
        self.show_photon_checkbox.stateChanged.connect(
            self._change_photon_show)
        self.show_grid_checkbox.stateChanged.connect(self._change_grid_show)
        self.show_hemisphere_checkbox.stateChanged.connect(
            self._change_hemisphere_show)
        self.show_isosurface_checkbox.stateChanged.connect(
            self._change_mesh_show)
        self.iso_spinbox.valueChanged.connect(self._change_iso_val)



"""Defines the crosshair class.

This file defines a class named crosshair to be the model side of the
crosshair. Addtitionally to a bare bones crosshair class, there is also
a corsshair class including a ROI and a ROI-annulus.
"""

import numpy as np
import copy
from kmap.library.misc import (
    idx_closest_value, centered_meshgrid, distance_in_meshgrid)


class CrosshairModel():

    def __init__(self, x=0, y=0):

        self.x = x
        self.y = y

    def set_position(self, x=None, y=None):

        if x is not None:
            self.x = x

        if y is not None:
            self.y = y

    def mask(self, plotdata, region='center', inverted=False):

        mask = np.zeros(plotdata.data.shape)

        x_idx = idx_closest_value(plotdata.x_axis, self.x)
        y_idx = idx_closest_value(plotdata.y_axis, self.y)
        
        if region == 'center':
            if x_idx is not None and y_idx is not None:
                mask[y_idx, x_idx] = 1

        elif region == 'x':
            if x_idx is not None:
                mask[:, x_idx] = 1

        elif region == 'y':
            if y_idx is not None:
                mask[y_idx, :] = 1

        mask = mask.astype(np.bool, copy=False)

        if inverted:
            return ~mask

        else:
            return mask

    def cut_from_data(self, plotdata, fill=np.nan, *args, **kwargs):

        mask = self.mask(plotdata, *args, **kwargs)
        cut_data = copy.deepcopy(plotdata)
        cut_data.data[~mask] = fill

        return cut_data


class CrosshairROIModel(CrosshairModel):

    def __init__(self, x=0, y=0, radius=0):

        if not np.isfinite(radius) or not radius >= 0:
            raise ValueError('radius has to be positive finite')
        else:
            self.radius = radius

        super().__init__(x, y)

    def set_radius(self, radius):

        if not np.isfinite(radius) or not radius >= 0:
            raise ValueError('radius has to be positive finite')
        else:
            self.radius = radius

    def mask(self, plotdata, region='center', inverted=False):

        if region in ['center', 'x', 'y']:
            mask = super().mask(plotdata, region=region, inverted=inverted)

        else:
            # Mesgrid of axis values shifted by center
            X, Y = centered_meshgrid(
                plotdata.x_axis, self.x, plotdata.y_axis, self.y)

            if region == 'roi':
                # Calculate distance from center for all values
                distance = distance_in_meshgrid(Y, X)
                mask = np.array(distance < self.radius, dtype=np.bool)

            # region == 'border'
            else:
                # "Pixel" is on the border if the border radius lies
                # inside (strictly) the pixels max and min radius

                epsilon = plotdata.step_size / 2

                sign_y = np.sign(Y)
                sign_x = np.sign(X)
                lower_bound = distance_in_meshgrid(
                    Y - sign_y * epsilon[1], X - sign_x * epsilon[0])

                sign_y[sign_y == 0] = 1
                sign_x[sign_x == 0] = 1
                upper_bound = distance_in_meshgrid(
                    Y + sign_y * epsilon[1], X + sign_x * epsilon[0])

                mask = np.array(
                    (lower_bound < self.radius) & (
                        self.radius < upper_bound),
                    dtype=np.bool)

            if inverted:
                mask = ~mask

        return mask


class CrosshairAnnulusModel(CrosshairROIModel):

    def __init__(self, x=0, y=0, radius=0, width=0):

        if not np.isfinite(width) or not width >= 0:
            raise ValueError('width has to be positive finite')
        else:
            self.width = width

        super().__init__(x, y, radius)

    def set_width(self, width):

        if not np.isfinite(width) or not width >= 0:
            raise ValueError('width has to be positive finite')
        else:
            self.width = width

    def mask(self, plotdata, region='center', inverted=False):

        if region in ['center', 'x', 'y', 'roi', 'border']:
            mask = super().mask(plotdata, region=region, inverted=inverted)

        else:
            # Utilize the CrosshairROI class
            aux_crosshair = CrosshairROIModel(
                x=self.x, y=self.y, radius=self.radius + self.width)

            if region == 'outer_border':
                mask = aux_crosshair.mask(
                    plotdata, region='border', inverted=inverted)

            # region == 'ring'
            else:
                mask = aux_crosshair.mask(plotdata, region='roi')
                aux_crosshair.set_radius(self.radius)
                inner_mask = aux_crosshair.mask(plotdata, region='roi')

                mask[inner_mask] = False

                if inverted:
                    mask = ~mask

        return mask



from setuptools import setup, find_packages
from kmap import __version__

setup(name='kMap',
      version=__version__,
      description='Display, modify and analyse ARPES experiments and DFT \
      simulations.',
      packages=find_packages(),
      license='MIT',
      url='https://github.com/brands-d/kMap',
      author='Dominik Brandstetter',
      author_email='dominik.brandstetter@edu.uni-graz.at',
      zip_safe=False,
      install_requires=['h5py==2.10.0', 'matplotlib==3.3.0', 'numpy==1.19.0',
                        'PyQt5==5.15.0', 'scipy==1.5.1', 'PyOpenGL==3.1.5',
                        'lmfit==1.0.1'],
      dependency_links=[
          'pyqtgraph git+https://github.com/pyqtgraph/pyqtgraph/tarball'
          '/abfac52c34e368171db472efc8efc581b6383f0d#egg=pyqtgraph'],
      include_package_data=True)



"""Unittests for the methods in the misc.py file.
"""

# Python Imports
import unittest

# Third Party Imports
import numpy.testing as npt

# Own Imports
from kmap.library.misc import *


class TestMisc(unittest.TestCase):

    def test_round_to(self):

        self.assertEqual(round_to(2, 0.25), 2)
        self.assertEqual(round_to(2.2, 0.25), 2.25)
        self.assertEqual(round_to(-1.1, 0.25), -1)

    def test_idx_closest_value(self):

        axis = [-1.73, -1.33, -0.93]

        value = -1.73
        self.assertEqual(idx_closest_value(axis, value), 0)

        value = -1.63
        self.assertEqual(idx_closest_value(axis, value), 0)

        value = 1.73
        self.assertEqual(idx_closest_value(axis, value), None)

        value = 1.73
        self.assertEqual(idx_closest_value(axis, value, bounds_error=False), 2)

        value = -1.94
        self.assertEqual(idx_closest_value(axis, value), None)

        value = -1.94
        self.assertEqual(idx_closest_value(axis, value, bounds_error=False), 0)

        value = -1.23
        self.assertEqual(idx_closest_value(axis, value), 1)

    def test_centered_meshgrid(self):

        x = np.array([1, 2, 3])
        y = np.array([-1, 0, 1])
        expected = ([[-1, 0, 1]], [[-1], [0], [1]])

        npt.assert_equal(centered_meshgrid(x, 2, y, 0), expected)

    def distance_in_meshgrid(self):

        X, Y = [[-1, 0, 1]], [[-1], [0], [1]]
        sqrt2 = np.sqrt(2)
        expected = [[sqrt2, 1, sqrt2], [1, 0, 1], [sqrt2, 1, sqrt2]]

        npt.assert_almost_equal(distance_in_meshgrid(X, Y), expected)

    def test_normalize(self):

        data = [1, 2, 3]

        self.assertEqual(normalize(data), 2)

    def test_axis_from_range(self):

        npt.assert_equal(axis_from_range([1, 2], 3), [1, 1.5, 2])

    def test_range_from_axes(self):

        x = [-1, 0, 1]
        y = [0, 2, 4]
        expected_ranges = [[-1, 1], [0, 4]]
        expected_step_sizes = [1, 2]

        npt.assert_equal(range_from_axes(x, y), (
                         expected_ranges, expected_step_sizes))

    def test_step_size_to_num(self):

        npt.assert_equal(step_size_to_num([1, 6], 2), 3)
        npt.assert_equal(step_size_to_num([-1, 1], 1), 3)

    def test_get_rotation_axes(self):

        sqrt1_2 = 1 / np.sqrt(2)
        expected = [[0, 0, 1], [sqrt1_2, sqrt1_2, 0], [sqrt1_2, -sqrt1_2, 0]]

        npt.assert_almost_equal(get_rotation_axes(45, 90), expected)

    def test_get_reduced_chi2(self):

        npt.assert_almost_equal(get_reduced_chi2([1, 2, 3], 2), 14)



"""Defines the LMFitModel class.

This file defines a class named LMFitModel designed encapsulate the
LMFit process. As such it can be understood as a wrapper class for the
lmfit package, both wrapping it's methods into a more useful form and
extending the feature-set. This class is can be used as a module inside
scripts without the GUI part.
"""

# Python Imports
import copy
import re
import builtins

# Third Party Imports
import numpy as np
from lmfit import minimize, Parameters

# Own Imports
from kmap.model.crosshair_model import CrosshairModel, CrosshairAnnulusModel
from kmap.library.orbitaldata import OrbitalData
from kmap.library.sliceddata import SlicedData
from kmap.library.misc import (
    axis_from_range, step_size_to_num, get_reduced_chi2, transpose_axis_order)
from kmap.config.config import config


class LMFitModel():
    """
    Wrapper class for the lmfit package. Acts both as module usable in
    scripts as well as core logic class for the lmfit part in kMap.py.

    For an example on how to use it please see the 'test_PTCDA' test in
    the 'kmap.tests.test_lmfit' file.

    ATTENTION: Please do not set any attributes manually. Instead use
    the appropriate "set_xxx" method instead.
    """

    def __init__(self, sliced_data, orbitals):
        """
        Args:
            sliced_data (SlicedData): A single SlicedData object.
            orbitals (OrbitalData or list): A single OrbitalData object
                or a list of OrbitalData objects.
                ATTENTION: An Orbital object is NOT sufficient. Please
                use the OrbitalData wrapping class instead.
        """

        self.axis = None
        self.crosshair = None
        self.symmetrization = 'no'
        self.background_equation = ['0', []]
        self.Ak_type = 'no'
        self.polarization = 'p'
        self.slice_policy = [0, [0], False]
        self.method = {'method': 'leastsq', 'xtol': 1e-12}
        self.region = ['all', False]

        self._set_sliced_data(sliced_data)
        self._add_orbitals(orbitals)

        self._set_parameters()

    def set_crosshair(self, crosshair):
        """A setter method to set a custom crosshair. If none is set
        when a region restriction is applied, a CrosshairAnnulusModel
        will be created.

        Args:
            crosshair (CrosshairModel): A crosshair model for cutting
                the data for any region-restriction.
                ATTENTION: The passed CrosshairModel has to support the
                region restriction you want to use.
        """

        if crosshair is None or isinstance(crosshair, CrosshairModel):
            self.crosshair = crosshair

        else:
            raise TypeError(
                'crosshair has to be of type %s (is %s)' % (
                    type(CrosshairModel), type(crosshair)))

    def set_axis(self, axis):
        """A setter method to set an axis for the interpolation onto a
        common grid. Default is the x-axis of the first slice in the
        list of slices chosen to be fitted.

        Args:
            axis (np.array): 1D array defining the common axis (and grid
            as only square kmaps are supported) for the subtraction.
        """

        self.axis = axis

    def set_axis_by_step_size(self, range_, step_size):
        """A convenience setter method to set an axis by defining the
        range and the step size.

        Args:
            range_ (list): A list of min and max value.
            step_size (float): A number denoting the step size.
        """

        num = step_size_to_num(range_, step_size)
        self.set_axis(axis_from_range(range_, num))

    def set_axis_by_num(self, range_, num):
        """A convenience setter method to set an axis by defining the
        range and the number of grid points.

        Args:
            range_ (list): A list of min and max value.
            num (int): An integer denoting the number of grid points.
        """

        self.set_axis(axis_from_range(range_, num))

    def set_symmetrization(self, symmetrization):
        """A setter method to set the type of symmetrization for the
        orbital kmaps. Default is 'no'.

        Args:
            symmetrization (str): See 'get_kmap' from
            'kmap.library.orbital.py' for information.
        """

        self.symmetrization = symmetrization

    def set_region(self, region, inverted=False):
        """A setter method to set the region restriction for the lmfit
        process. Default is no region restriction ('all').

        Args:
            region (str): Supports all regions the crosshair model you
            supplied supports. See there for documentation. (default
            is a CrosshairAnnulusModel).
            inverted (bool): See your CrosshairModel for documentation.
        """

        self.region = [region, inverted]

        if region != 'all' and self.crosshair is None:
            self.crosshair = CrosshairAnnulusModel()

    def set_polarization(self, Ak_type, polarization):
        """A setter method to set the type of polarization for the
        orbital kmaps. Default is 'toroid' and 'p'.

        Args:
            Ak_type (str): See 'get_kmap' from
            'kmap.library.orbital.py' for information.
            polarization (str): See 'get_kmap' from
            'kmap.library.orbital.py' for information.
        """

        self.Ak_type = Ak_type
        self.polarization = polarization

    def set_slices(self, slice_indices, axis_index=0, combined=False):
        """A setter method to chose the slices to be fitted next time
        'fit()' is called. Default is [0], 0 and False.

        Args:
            slice_indices (int or list or str): Either one or more
            indices for the slices to be fitted next. Pass 'all' to use
            all slices in this axis.
            axis_index (int): Which axis in the SlicedData is used as
            slice axis.
            combined (bool): Whether to fit all slices individually or
            add all the slices for one fit instead.
        """

        if isinstance(slice_indices, str) and slice_indices == 'all':
            self.slice_policy = [axis_index,
                                 range(self.sliced_data.axes[axis_index].num),
                                 combined]

        elif isinstance(slice_indices, list):
            self.slice_policy = [axis_index, slice_indices, combined]

        elif isinstance(slice_indices, range):
            self.slice_policy = [axis_index, list(slice_indices), combined]

        else:
            self.slice_policy = [axis_index, [slice_indices], combined]

    def set_fit_method(self, method, xtol=1e-7):
        """A setter method to set the method and the tolerance for the
        fitting process. Default is 'leastsq' and 1e-7.

        Args:
            method (str): See the documentation for the lmfit module.
            xtol (str): Only available for one of the following methods:
                        'leastsq', 'least_squares', 'powell'
        """

        xtol_capable_methods = ['leastsq', 'least_squares']
        if method in xtol_capable_methods:
            self.method = {'method': method, 'xtol': xtol}

        else:
            self.method = {'method': method}

    def set_background_equation(self, equation):
        """A setter method to set an custom background equation.
        Default is '1'.

        Args:
            equation (str): An equation used to calculate the background
            profile. Can use python function (e.g. abs()) and basics
            methods from the numpy module (prefix by 'np.';
            e.g. np.sqrt()). Can contain variables to be fitted.
            Variables have can only contain lower or upper case letters,
            underscores and numbers. They cannot start with numbers.
            The variables 'x' and 'y' are special and denote the x and
            y axis respectively. No variables already used outside the
            background equation (like phi) can be used.
            Here are some examples of valid variable names:
            x_s, x2, x_2, foo, this_is_a_valid_variable.
            Each variables starts with following default values:
            value=0, min=-99999.9, max=99999.9, vary=False, expr=None

            The equation will be parsed by eval. Please don't injected
            any code as it would be really easy to do so. There are no
            safeguards in place whatsoever so we (have to) trust you.
            Thanks, D.B.
        """

        try:
            compile(equation, '', 'exec')

        except:
            raise ValueError(
                'Equation is not parseable. Check for syntax errors.')

        # Pattern matches all numpy, math and builtin methods
        clean_pattern = 'np\\.[a-z1-9\\_]+|math\\.[a-z1-9\\_]+'
        for builtin in dir(builtins):
            clean_pattern += '|' + str(builtin)

        cleaned_equation = re.sub(clean_pattern, '', equation)
        # Pattern matches all text including optional underscore with
        # numbers.
        variable_pattern = '[a-zA-Z\\_]+[0-9]*'
        variables = list(set(re.findall(variable_pattern, cleaned_equation)))

        # x and y need special treatment
        if 'x' in variables:
            variables.remove('x')

        if 'y' in variables:
            variables.remove('y')

        new_variables = np.setdiff1d(variables, self.background_equation[1])
        self.background_equation = [equation, variables]
        for variable in new_variables:
            self.parameters.add(variable, value=0, min=-99999.9,
                                max=99999.9, vary=False, expr=None)

        return [self.parameters[variable] for variable in new_variables]

    def edit_parameter(self, parameter, *args, **kwargs):
        """A setter method to edit fitting settings for one parameter.
        Use this method to enable a parameter for fitting (vary=True)

        Args:
            parameter (str): Name of the parameter to be editted.
            *args & **kwargs (): Are being passed to the
            'parameter.set' method of the lmfit module. See there
            for more documentation.
        """

        self.parameters[parameter].set(*args, **kwargs)

    def fit(self):
        """Calling this method will trigger a lmfit with the current
        settings.

        Returns:
            (list): A list of MinimizerResults. One for each slice
            fitted.
        """

        lmfit_padding = float(config.get_key('lmfit', 'padding'))

        any_parameter_vary = False
        for parameter in self.parameters.values():
            if parameter.vary:
                any_parameter_vary = True
                break

        if not any_parameter_vary and self.method['method'] not in ['leastsq', 'least_squares']:
            raise ValueError(
                'Only leastsq and least_squares can fit if no parameter is set to vary.')

        for parameter in self.parameters.values():
            if parameter.vary and parameter.value <= parameter.min:
                padded_value = parameter.min + lmfit_padding
                print('WARNING: Initial value for parameter \'%s\' had to be corrected to %f (was %f)' % (
                    parameter.name, padded_value, parameter.value))
                parameter.value = padded_value

        results = []
        for index in self.slice_policy[1]:
            slice_ = self.get_sliced_kmap(index)
            result = minimize(self._chi2,
                              copy.deepcopy(self.parameters),
                              kws={'slice_': slice_},
                              nan_policy='omit',
                              **self.method)

            results.append([index, result])

        return results

    def transpose(self, constant_axis):

        axis_order = transpose_axis_order(constant_axis)

        self.sliced_data.transpose(axis_order)

    def get_settings(self):

        settings = {'crosshair': self.crosshair,
                    'background': self.background_equation,
                    'symmetrization': self.symmetrization,
                    'polarization': [self.Ak_type, self.polarization],
                    'slice_policy': self.slice_policy,
                    'method': self.method,
                    'region': self.region,
                    'axis': self.axis}

        return copy.deepcopy(settings)

    def set_settings(self, settings):

        self.set_crosshair(settings['crosshair'])
        self.set_background_equation(settings['background'][0])
        self.set_polarization(*settings['polarization'])
        slice_policy = settings['slice_policy']
        self.set_slices(slice_policy[1], slice_policy[0], slice_policy[2])
        self.set_region(*settings['region'])
        self.set_symmetrization(settings['symmetrization'])
        self.set_fit_method(*settings['method'])
        self.set_axis(settings['axis'])

    def get_sliced_kmap(self, slice_index):

        axis_index, slice_indices, is_combined = self.slice_policy

        if is_combined:
            kmaps = []
            for slice_index in slice_indices:
                kmaps.append(self.sliced_data.slice_from_index(slice_index,
                                                               axis_index))

            kmap = np.nansum(kmaps, axis=axis_index)

        else:
            kmap = self.sliced_data.slice_from_index(slice_index,
                                                     axis_index)

        if self.axis is not None:
            kmap = kmap.interpolate(self.axis, self.axis)

        else:
            self.axis = kmap.x_axis

        return kmap

    def get_orbital_kmap(self, ID, param=None):

        if param is None:
            param = self.parameters

        orbital = self.ID_to_orbital(ID)
        kmap = orbital.get_kmap(E_kin=param['E_kin'].value,
                                dk=(self.axis, self.axis),
                                phi=param['phi_' + str(ID)].value,
                                theta=param['theta_' + str(ID)].value,
                                psi=param['psi_' + str(ID)].value,
                                alpha=param['alpha'].value,
                                beta=param['beta'].value,
                                Ak_type=self.Ak_type,
                                polarization=self.polarization,
                                symmetrization=self.symmetrization)

        return kmap

    def get_weighted_sum_kmap(self, param=None, with_background=True):

        if param is None:
            param = self.parameters

        orbital_kmaps = []
        for orbital in self.orbitals:
            ID = orbital.ID

            weight = param['w_' + str(ID)].value

            kmap = weight * self.get_orbital_kmap(ID, param)

            orbital_kmaps.append(kmap)

        orbital_kmap = np.nansum(orbital_kmaps)

        if with_background:
            variables = {}
            for variable in self.background_equation[1]:
                variables.update({variable: param[variable].value})

            background = self._get_background(variables)

            return orbital_kmap + background

        else:
            return orbital_kmap

    def get_residual(self, slice_, param=None, weight_sum_data=None):

        if param is None:
            param = self.parameters

        if weight_sum_data is None:
            orbital_kmap = self.get_weighted_sum_kmap(param)

        else:
            orbital_kmap = weight_sum_data

        if isinstance(slice_, int):
            sliced_kmap = self.get_sliced_kmap(slice_)
            residual = sliced_kmap - orbital_kmap

        else:
            residual = slice_ - orbital_kmap

        residual = self._cut_region(residual)

        return residual

    def get_reduced_chi2(self, slice_index, weight_sum_data=None):

        n = self._get_degrees_of_freedom()
        residual = self.get_residual(
            slice_index, weight_sum_data=weight_sum_data)
        reduced_chi2 = get_reduced_chi2(residual.data, n)

        return reduced_chi2

    def ID_to_orbital(self, ID):

        for orbital in self.orbitals:
            if orbital.ID == ID:
                return orbital

        return None

    def _chi2(self, param=None, slice_=0):

        if param is None:
            param = self.parameters

        residual = self.get_residual(slice_, param)

        return residual.data

    def _get_degrees_of_freedom(self):

        n = 0
        for parameter in self.parameters.values():
            if parameter.vary:
                n += 1

        return n

    def _get_background(self, variables=[]):

        variables.update({'x': self.axis, 'y': np.array([self.axis]).T})
        background = eval(self.background_equation[0], None, variables)

        return background

    def _cut_region(self, data):

        if self.crosshair is None or self.region[0] == 'all':
            return data

        else:
            return self.crosshair.cut_from_data(
                data, region=self.region[0], inverted=self.region[1])

    def _set_sliced_data(self, sliced_data):

        if isinstance(sliced_data, SlicedData):
            self.sliced_data = sliced_data

        else:
            raise TypeError(
                'sliced_data has to be of type %s (is %s)' % (
                    type(SlicedData), type(sliced_data)))

    def _add_orbitals(self, orbitals):

        if (isinstance(orbitals, list) and
                all(isinstance(element, OrbitalData)
                    for element in orbitals)):
            self.orbitals = orbitals

        elif isinstance(orbitals, OrbitalData):
            self.orbitals = [orbitals]

        else:
            raise TypeError(
                'orbital has to be of (or list of) type %s (is %s)' % (
                    type(OrbitalData), type(orbitals)))

    def _set_parameters(self):

        self.parameters = Parameters()

        for orbital in self.orbitals:
            ID = orbital.ID

            self.parameters.add('w_' + str(ID), value=1,
                                min=0, vary=False, expr=None)
            for angle in ['phi_', 'theta_', 'psi_']:
                self.parameters.add(angle + str(ID), value=0,
                                    min=90, max=-90, vary=False, expr=None)

        # LMFit doesn't work when the initial value is exactly the same
        # as the minimum value. For this reason the initial value will
        # be set ever so slightly above 0 to circumvent this problem.
        self.parameters.add('c', value=0,
                            min=0, vary=False, expr=None)
        self.parameters.add('E_kin', value=30,
                            min=5, max=150, vary=False, expr=None)
        for angle in ['alpha', 'beta']:
            self.parameters.add(angle, value=0,
                                min=90, max=-90, vary=False, expr=None)



# Python Imports
import logging
import traceback

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtGui import QIcon, QKeySequence
from PyQt5.QtWidgets import QMainWindow, QFileDialog, QMessageBox
from PyQt5.QtCore import QDir

# Own Imports
from kmap import __directory__, __version__
from kmap.controller.databasewindows import (
    OrbitalDatabase, SlicedDatabaseWindow)
from kmap.controller.tabwidget import TabWidget
from kmap.controller.tabchoosewindow import TabChooseWindow
from kmap.model.mainwindow_model import MainWindowModel
from kmap.controller.sliceddatatab import SlicedDataTab
from kmap.controller.orbitaldatatab import OrbitalDataTab
from kmap.config.config import config

# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/mainwindow.ui')
MainWindow_UI, _ = uic.loadUiType(UI_file)


class MainWindow(QMainWindow, MainWindow_UI):

    def __init__(self):

        # Setup GUI
        super(MainWindow, self).__init__()
        self.setupUi(self)
        self._setup()
        self._connect()

        self.sub_windows = {}

        self.model = MainWindowModel(self)

        self.open_welcome()

    def load_hdf5_files(self):
        # Load one or more new hdf5 files

        log = logging.getLogger('kmap')
        log.info('Loading .hdf5 file(s)...')

        start_path = __directory__ + config.get_key('paths', 'hdf5_start')
        extensions = 'hdf5 files (*.hdf5 *.h5);; All Files (*)'
        paths, _ = QFileDialog.getOpenFileNames(
            None, 'Open file(s)', start_path, extensions)

        if not paths:
            # No file chosen
            log.info('No file chosen')
            return

        for path in paths:

            self.tab_widget.open_sliced_data_tab(path)

    def open_orbitaldatabase_browser(self):

        database = OrbitalDatabase()
        database.files_chosen.connect(self.load_cube_files_online)

        self.sub_windows.update({str(id(database)): database})

    def open_binding_energy_sliceddatabase_browser(self):

        database = SlicedDatabaseWindow(mode='binding-energy')
        database.files_chosen.connect(self.load_sliced_files_online)

        self.sub_windows.update({str(id(database)): database})

    def open_photon_energy_sliceddatabase_browser(self):

        database = SlicedDatabaseWindow(mode='photon-energy')
        database.files_chosen.connect(self.load_sliced_file_online)

        self.sub_windows.update({str(id(database)): database})

    def open_cubefile_sliceddatabase_browser(self):

        database = SlicedDatabaseWindow(mode='cubefile')
        database.files_chosen.connect(self.load_cubefile_as_sliced_online)

        self.sub_windows.update({str(id(database)): database})

    def load_sliced_files_online(self, URLs):
        # Load one or more cube files as sliced data[BE, kx, ky]

        log = logging.getLogger('kmap')
        log.info('Loading .cube file(s) as sliced data[BE, kx, ky]...')

        self.tab_widget.open_sliced_data_tab_by_URLs(URLs)

    def load_sliced_file_online(self, URL):
        # Load one cube file as sliced data[photon_energy, kx, ky]

        log = logging.getLogger('kmap')
        log.info('Loading .cube file as sliced data[photon_energy, kx, ky]...')

        self.tab_widget.open_sliced_data_tab_by_URL(URL)

    def load_cubefile_as_sliced_online(self, URL):
        # Load one cube file as sliced psi[x,y,z] or psik[kx,ky,kz] 

        log = logging.getLogger('kmap')
        log.info('Loading .cube file as sliced psi[x,y,z] or psik[kx,ky,kz] ...')

        self.tab_widget.open_sliced_data_tab_by_cube(URL)

    def load_cube_files_online(self, URLs):

        if not URLs:
            # No file chosen
            logging.getLogger('kmap').info('No file chosen')
            return

        # Get Tab to load to
        tab = self.tab_widget.get_orbital_tab_to_load_to()

        # Load URL to Tab
        for URL in URLs:

            try:
                tab.add_orbital_from_online(URL[0], URL[1])

            except ValueError:
                logging.getLogger('kmap').error(
                    'URL "%s" is invalid. File could not be loaded.' % URL)

    def load_cube_files_locally(self):
        # Load one or more new cube files

        start_path = __directory__ + config.get_key('paths', 'cube_start')
        extensions = 'cube files (*.cube);; All Files (*)'
        paths, _ = QFileDialog.getOpenFileNames(
            None, 'Open file(s)', start_path, extensions)

        if not paths:
            # No file chosen
            logging.getLogger('kmap').info('No file chosen')
            return

        # Get Tab to load to
        tab = self.tab_widget.get_orbital_tab_to_load_to()

        # Load Data to Tab
        for path in paths:
            tab.add_orbital_from_filepath(path)

    def open_log_file(self):
        # Open log file

        path = __directory__ + QDir.toNativeSeparators('/../default.log')
        title = 'Log File'

        self.tab_widget.open_file_tab(
            path, title, editable=False, richText=False)

    def open_mod_log_file(self):
        # Open module log file

        path = __directory__ + QDir.toNativeSeparators('/../modules.log')
        title = 'Modules Log File'

        self.tab_widget.open_file_tab(
            path, title, editable=False, richText=False)

    def open_general_settings(self):
        # Open general user settings

        path = __directory__ + \
            QDir.toNativeSeparators('/config/settings_user.ini')
        title = 'General Settings'

        self.tab_widget.open_file_tab(
            path, title, editable=True, richText=False)

    def open_orbital_data_tab(self):

        self.tab_widget.open_orbital_data_tab()

    def open_profile_tab(self):

        self.tab_widget.open_profile_tab()

    def open_logging_settings(self):
        # Open logging user settings

        path = __directory__ + \
            QDir.toNativeSeparators('/config/logging_user.ini')
        title = 'Logging Settings'

        self.tab_widget.open_file_tab(
            path, title, editable=True, richText=False)

    def open_general_default_settings(self):
        # Open general default settings

        path = __directory__ + \
            QDir.toNativeSeparators('/config/settings_default.ini')
        title = 'General Settings (Default)'

        self.tab_widget.open_file_tab(
            path, title, editable=False, richText=False)

    def open_logging_default_settings(self):
        # Open logging default settings

        path = __directory__ + \
            QDir.toNativeSeparators('/config/logging_default.ini')
        title = 'Logging Settings (Default)'

        self.tab_widget.open_file_tab(
            path, title, editable=False, richText=False)

    def open_readme(self):
        # Open a README page

        path = __directory__ + \
            QDir.toNativeSeparators('/resources/texts/readme.txt')
        title = 'README'

        self.tab_widget.open_file_tab(
            path, title, editable=False, richText=True)

    def open_welcome(self):
        # Open a welcome page

        path = __directory__ + \
            QDir.toNativeSeparators('/resources/texts/welcome.txt')
        title = 'Welcome'

        self.tab_widget.open_file_tab(
            path, title, editable=False, richText=True)

    def open_about(self):
        # Open an about window

        path = __directory__ + \
            QDir.toNativeSeparators('/resources/texts/about.txt')
        title, text = self.model.get_about_text(path)
        QMessageBox.about(self, title, text)

    def open_lmfit_tab(self, tabs):

        if len(tabs) != 2 or tabs[0] is None or tabs[1] is None:
            return

        self.tab_widget.open_lmfit_tab(*tabs)

    def open_tab_choose_window(self):

        sliced_tabs = self.tab_widget.get_tabs_of_type(SlicedDataTab)
        orbital_tabs = self.tab_widget.get_tabs_of_type(OrbitalDataTab)

        if len(sliced_tabs) == 0 or len(orbital_tabs) == 0:
            return

        elif len(sliced_tabs) == 1 and len(orbital_tabs) == 1:
            tabs = [*sliced_tabs, *orbital_tabs]
            self.open_lmfit_tab(tabs)

        else:
            self.tab_choose_window = TabChooseWindow(sliced_tabs,
                                                     orbital_tabs)
            self.tab_choose_window.tabs_chosen.connect(self.open_lmfit_tab)

    def open_in_matplotlib(self):

        current_tab = self.tab_widget.get_current_tab()

        if hasattr(current_tab, 'display_in_matplotlib'):
            window = current_tab.display_in_matplotlib()
            self.sub_windows.update({str(id(window)): window})

    def export_to_txt(self):

        current_tab = self.tab_widget.get_current_tab()

        if hasattr(current_tab, 'export_to_txt'):
            text = current_tab.export_to_txt()

            file_name, _ = QFileDialog.getSaveFileName(self, 'Save .txt File')
            print(file_name)
            with open(file_name, 'w') as file:
                file.write(text)

    def reload_settings(self):
        # Reload the settings

        config.setup()

    def closeEvent(self, event):

        for window in self.sub_windows.values():
            try:
                window.close()

            except Exception as e:
                pass

        event.accept()

    def _setup(self):

        self._set_misc()
        self._initialize_shortcuts()

    def _set_misc(self):

        x = int(config.get_key('app', 'x'))
        y = int(config.get_key('app', 'y'))
        w = int(config.get_key('app', 'w'))
        h = int(config.get_key('app', 'h'))
        self.setGeometry(x, y, h, w)
        if config.get_key('app', 'fullscreen') == 'True':
            self.showMaximized()
        self.setWindowTitle('kMap.py')
        self.setWindowIcon(QIcon(__directory__ +
                                 QDir.toNativeSeparators(
                                     '/resources/images/icon.png')))
        self.show()

    def _initialize_shortcuts(self):

        actions = [self.load_hdf5_action,
                   self.load_sliced_from_binding_energy_action,
                   self.load_sliced_from_photon_energy_action,
                   self.show_matplotlib,
                   self.log_file_action, self.load_cube_online_action,
                   self.load_cube_file_action,
                   self.open_lmfit_tab_action, self.settings_action]

        alias = ['load_hdf5', 'load_sliced_from_binding_energy_action',
                 'load_sliced_from_photon_energy_action',
                 'show_matplotlib', 'open_log',
                 'load_cube_online', 'load_cube_file',
                 'open_lmfit', 'reload_settings']

        for action, alias in zip(actions, alias):
            shortcut = config.get_key('shortcut', alias, file='shortcut')
            action.setShortcut(QKeySequence(shortcut))

        self.ref_action.setShortcut(QKeySequence('Ctrl+r'))

    def _connect(self):

        # File menu
        self.load_hdf5_action.triggered.connect(self.load_hdf5_files)
        self.load_sliced_from_binding_energy_action.triggered.connect(
            self.open_binding_energy_sliceddatabase_browser)
        self.load_sliced_from_photon_energy_action.triggered.connect(
            self.open_photon_energy_sliceddatabase_browser)
        self.load_sliced_from_cubefile_action.triggered.connect(
            self.open_cubefile_sliceddatabase_browser)        
        self.load_cube_file_action.triggered.connect(
            self.load_cube_files_locally)
        self.load_cube_online_action.triggered.connect(
            self.open_orbitaldatabase_browser)
        self.log_file_action.triggered.connect(self.open_log_file)
        self.mod_log_file_action.triggered.connect(self.open_mod_log_file)
        self.show_matplotlib.triggered.connect(self.open_in_matplotlib)
        self.export_txt.triggered.connect(self.export_to_txt)

        # Tabs menu
        self.open_sim_tab_action.triggered.connect(self.open_orbital_data_tab)
        self.open_profile_tab_action.triggered.connect(self.open_profile_tab)
        self.open_lmfit_tab_action.triggered.connect(
            self.open_tab_choose_window)

        # Preferences menu
        self.general_action.triggered.connect(self.open_general_settings)
        self.logging_action.triggered.connect(self.open_logging_settings)
        self.general_default_action.triggered.connect(
            self.open_general_default_settings)
        self.logging_default_action.triggered.connect(
            self.open_logging_default_settings)
        self.settings_action.triggered.connect(self.reload_settings)

        # Help menu
        self.readme_action.triggered.connect(self.open_readme)
        self.welcome_action.triggered.connect(self.open_welcome)
        self.about_action.triggered.connect(self.open_about)



# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import Qt, QDir

# Own Imports
from kmap import __directory__

# Load .ui File
UI_file = __directory__ + \
    QDir.toNativeSeparators('/ui/slicedcubefileoptions.ui')
SlicedDataBaseOptions_UI, _ = uic.loadUiType(UI_file)


class SlicedCubefileOptions(QWidget, SlicedDataBaseOptions_UI):

    def __init__(self, *args, **kwargs):

        # Setup GUI
        super(SlicedCubefileOptions, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._setup()

    def get_parameters(self):

        name = self.line_edit.text()
        name = name if name else 'no name given'
        E_kin_max = self.Ekin_max_spinbox.value()
        dk3D = self.dk3D_step_spinbox.value()
        domain_index = self.domain_combobox.currentIndex()
        value_index = self.value_combobox.currentIndex()

        if domain_index == 0:
            domain = 'real-space' 

        else:
            domain = 'k-space'

        if value_index == 0:   
            value = 'real'

        elif value_index == 1:
            value = 'imag'

        elif value_index == 2:
            value = 'abs'

        elif value_index == 3:
            value = 'abs2'


        return (name, domain, dk3D, E_kin_max, value)

    def closeEvent(self, event):

        self.deleteLater()
        event.accept()

    def _setup(self):

        self.setWindowFlags(Qt.CustomizeWindowHint | Qt.WindowTitleHint)



class AbstractData():

    def __init__(self, ID, name, meta_data):

        self.name = name
        self.ID = ID
        self.meta_data = meta_data

    def __str__(self):

        rep = 'Name:\t%s\nID:\t%s\n\nMeta Data\n' % (self.name, self.ID)

        for key, value in self.meta_data.items():
            rep += '%s:\t\t%s\n' % (key, str(value))

        # Don't pass last new line
        return rep[:-2]



import json
import os

import traceback
from pyqtgraph import ColorMap
from kmap.library.colormap import Colormap


class ColormapModel():

    def __init__(self, plot_item):

        self.colormaps = []

        if isinstance(plot_item, list):
            self.plot_item = plot_item

        else:
            self.plot_item = [plot_item]

    def load_colormaps(self, path):
        # Load colormaps from json file

        with open(path, 'r') as file:

            data = json.loads(json.load(file))
            self.colormaps = []

            for colormap in data:
                name = colormap[0]
                pos = colormap[1]
                colors = colormap[2]

                self.add_colormap(name, pos, colors)

    def add_colormap_from_plot(self, name, ID=0):
        # Add current plot_item colormap as new colormap

        colormap = self.plot_item[ID].ui.histogram.gradient.colorMap()
        pos = colormap.pos.tolist()
        colors = colormap.getColors().tolist()

        self.add_colormap(name, pos, colors)

    def add_colormap(self, name, pos, colors):
        # Add new colormap to list of colormaps

        new_colormap = Colormap(name, pos, colors)
        self.colormaps.append(new_colormap)

    def remove_colormap(self, name):

        index = self._name_to_index(name)

        del self.colormaps[index]

    def save_colormaps(self, path):
        # Save colormaps in json file

        path_temp = path + '.temp'

        with open(path_temp, 'w') as file:
            data = json.dumps([obj.toList() for obj in self.colormaps])
            json.dump(data, file)

        os.remove(path)
        os.rename(path_temp, path)

    def get_colormap(self, name):
        # Return a colormap by name

        for colormap in self.colormaps:
            if colormap.name == name:
                return colormap

        return None

    def set_current_colormap(self, name):
        # Change colormap of plot_item

        index = self._name_to_index(name)

        pos = self.colormaps[index].pos
        colors = self.colormaps[index].colors
        for plot_item in self.plot_item:
            plot_item.setColorMap(ColorMap(pos, colors))

    def _name_to_index(self, name):

        name_list = [colormap.name for colormap in self.colormaps]
        index = name_list.index(name)

        return index



# Python Imports
import logging

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import pyqtSignal, QDir
from PyQt5.QtWidgets import QWidget

# Own Imports
from kmap import __directory__

# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/polarization.ui')
Polarization_UI, _ = uic.loadUiType(UI_file)


class Polarization(QWidget, Polarization_UI):

    polarization_changed = pyqtSignal()

    def __init__(self, *args, **kwargs):

        # Setup GUI
        super(Polarization, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._connect()

    def change_polarization(self):

        self.polarization_changed.emit()

    def get_parameters(self):

        Ak_type, polarization = self._get_polarization()
        factor = self._get_factor()
        Ak_type = factor if factor == 'no' else factor + Ak_type
        angles = self._get_angles()

        return (Ak_type, polarization, *angles)

    def _get_factor(self):

        Ak_index = self.ak_combobox.currentIndex()

        if Ak_index == 0:
            Ak_type = 'no'

        elif Ak_index == 1:
            Ak_type = 'only-'

        else:
            Ak_type = ''

        return Ak_type

    def _get_polarization(self):

        Ak_index = self.polarization_combobox.currentIndex()

        if Ak_index == 0:
            Ak_type = 'toroid'
            polarization = 'p'

        else:
            Ak_type = 'NanoESCA'

            if Ak_index == 1:
                polarization = 'p'

            elif Ak_index == 2:
                polarization = 's'

            elif Ak_index == 3:
                polarization = 'unpolarized'                

            elif Ak_index == 4:
                polarization = 'C+'

            elif Ak_index == 5:
                polarization = 'C-'

            elif Ak_index == 6:
                polarization = 'CDAD'

        return Ak_type, polarization

    def _get_angles(self):

        alpha = self.angle_spinbox.value()
        beta = self.azimuth_spinbox.value()
        gamma = 'auto'

        return alpha, beta, gamma

    def _connect(self):

        self.polarization_combobox.currentIndexChanged.connect(
            self.change_polarization)
        self.angle_spinbox.valueChanged.connect(self.change_polarization)
        self.azimuth_spinbox.valueChanged.connect(self.change_polarization)
        self.ak_combobox.currentIndexChanged.connect(self.change_polarization)
